@@ -7,24 +7,28 @@
 #ifndef KANDROIDEXTRAS_JNIPROPERTIES_H
 #define KANDROIDEXTRAS_JNIPROPERTIES_H
 
-#include "jnicommon.h"
+#include "jniargument.h"
+#include "jniobject.h"
 #include "jnisignature.h"
 #include "jnitypes.h"
 #include "jnitypetraits.h"
 
-#include <QAndroidJniObject>
-
 #include <type_traits>
 
 namespace KAndroidExtras {
 
+namespace Jni {
+template <typename T> class Array;
+}
+
 /** @cond internal */
 namespace Internal {
 
 /** Wrapper for static properties. */
-template <typename PropType, typename ClassType, typename NameHolder, bool BasicType> struct StaticProperty {};
+template <typename PropType, typename ClassType, typename NameHolder, bool PrimitiveType> struct StaticProperty {};
 template <typename PropType, typename ClassType, typename NameHolder>
 struct StaticProperty<PropType, ClassType, NameHolder, false> {
+    static_assert(!is_invalid_primitive_type<PropType>::value, "Using an incompatible primitive type!");
     inline QAndroidJniObject get() const
     {
         return QAndroidJniObject::getStaticObjectField(Jni::typeName<ClassType>(), Jni::typeName<NameHolder>(), Jni::signature<PropType>());
@@ -37,13 +41,18 @@ struct StaticProperty<PropType, ClassType, NameHolder, false> {
     {
         return Jni::converter<PropType>::convert(get());
     }
+    template <typename RetT = PropType, typename = std::enable_if_t<Jni::is_generic_wrapper<PropType>::value, RetT>>
+    inline operator Jni::Object<PropType>() const
+    {
+        return Jni::Object<PropType>(get());
+    }
 };
 
 template <typename PropType, typename ClassType, typename NameHolder>
 struct StaticProperty<PropType, ClassType, NameHolder, true> {
-    inline operator PropType() const
+    inline operator auto() const
     {
-        return QAndroidJniObject::getStaticField<PropType>(Jni::typeName<ClassType>(), Jni::typeName<NameHolder>());
+        return primitive_value<PropType>::fromJni(QAndroidJniObject::getStaticField<typename primitive_value<PropType>::JniType>(Jni::typeName<ClassType>(), Jni::typeName<NameHolder>()));
     }
 };
 
@@ -53,16 +62,17 @@ class PropertyBase {
 protected:
     inline QAndroidJniObject handle() const {
         const auto owner = reinterpret_cast<const ClassType*>(reinterpret_cast<const char*>(this) - OffsetHolder::offset());
-        return owner->handle();
+        return owner->jniHandle();
     }
 };
 
 /** Wrapper for non-static properties. */
-template <typename PropType, typename ClassType, typename NameHolder, typename OffsetHolder, bool BasicType> struct Property {};
+template <typename PropType, typename ClassType, typename NameHolder, typename OffsetHolder, bool PrimitiveType> struct Property {};
 template <typename PropType, typename ClassType, typename NameHolder, typename OffsetHolder>
 class Property<PropType, ClassType, NameHolder, OffsetHolder, false> : public PropertyBase<ClassType, OffsetHolder> {
 private:
     struct _jni_NoType {};
+    static_assert(!is_invalid_primitive_type<PropType>::value, "Using an incompatible primitive type!");
 public:
     inline QAndroidJniObject get() const
     {
@@ -76,7 +86,22 @@ public:
     {
         return Jni::converter<PropType>::convert(get());
     }
+    template <typename RetT = PropType, typename = std::enable_if_t<Jni::is_array<PropType>::value, RetT>>
+    inline operator PropType() const
+    {
+        return PropType(get());
+    }
+    template <typename RetT = PropType, typename = std::enable_if_t<Jni::is_generic_wrapper<PropType>::value, RetT>>
+    inline operator Jni::Object<PropType>() const
+    {
+        return Jni::Object<PropType>(get());
+    }
 
+    inline Property& operator=(typename Internal::argument<PropType>::type value)
+    {
+        this->handle().setField(Jni::typeName<NameHolder>(), Jni::signature<PropType>(), Internal::argument<PropType>::toCallArgument(value));
+        return *this;
+    }
     inline Property& operator=(const QAndroidJniObject &value)
     {
         this->handle().setField(Jni::typeName<NameHolder>(), Jni::signature<PropType>(), value.object());
@@ -87,18 +112,26 @@ public:
         this->handle().setField(Jni::typeName<NameHolder>(), Jni::signature<PropType>(), Jni::reverse_converter<PropType>::type::convert(value).object());
         return *this;
     }
+
+    // special case for string comparison, which is often done against different types and thus the implicit conversion operator
+    // isn't going to be enough
+    template <typename CmpT, typename = std::enable_if_t<std::is_same_v<PropType, java::lang::String>, CmpT>>
+    inline bool operator==(const CmpT &other) const
+    {
+        return QString(*this) == other;
+    }
 };
 
 template <typename PropType, typename ClassType, typename NameHolder, typename OffsetHolder>
 class Property<PropType, ClassType, NameHolder, OffsetHolder, true> : public PropertyBase<ClassType, OffsetHolder> {
 public:
-    inline operator PropType() const
+    inline operator auto() const
     {
-        return this->handle().template getField<PropType>(Jni::typeName<NameHolder>());
+        return primitive_value<PropType>::fromJni(this->handle().template getField<typename primitive_value<PropType>::JniType>(Jni::typeName<NameHolder>()));
     }
     inline Property& operator=(PropType value)
     {
-        this->handle().setField(Jni::typeName<NameHolder>(), Jni::signature<PropType>(), value);
+        this->handle().setField(Jni::typeName<NameHolder>(), primitive_value<PropType>::toJni(value));
         return *this;
     }
 };
@@ -125,13 +158,13 @@ public:
  * copy the wrapper type, not read the property value.
  *
  * @param type The data type of the property.
- * @param name The name of the proeprty.
+ * @param name The name of the property.
  */
 #define JNI_CONSTANT(type, name) \
 private: \
     struct _jni_ ## name ## __NameHolder { static constexpr const char* jniName() { return "" #name; } }; \
 public: \
-    static inline const KAndroidExtras::Internal::StaticProperty<type, _jni_ThisType, _jni_ ## name ## __NameHolder, Jni::is_basic_type<type>::value> name;
+    static inline const KAndroidExtras::Internal::StaticProperty<type, _jni_ThisType, _jni_ ## name ## __NameHolder, Jni::is_primitive_type<type>::value> name;
 
 /**
  * Wrap a member property.
@@ -144,7 +177,7 @@ public: \
  * copy the wrapper type, not read the property value.
  *
  * @param type The data type of the property.
- * @param name The name of the proeprty.
+ * @param name The name of the property.
  */
 #define JNI_PROPERTY(type, name) \
 private: \
@@ -158,7 +191,7 @@ private: \
     }; \
     friend class KAndroidExtras::Internal::PropertyBase<_jni_ThisType, _jni_ ## name ## __OffsetHolder>; \
 public: \
-    [[no_unique_address]] KAndroidExtras::Internal::Property<type, _jni_ThisType, _jni_ ## name ## __NameHolder, _jni_ ## name ## __OffsetHolder, Jni::is_basic_type<type>::value> name;
+    [[no_unique_address]] KAndroidExtras::Internal::Property<type, _jni_ThisType, _jni_ ## name ## __NameHolder, _jni_ ## name ## __OffsetHolder, KAndroidExtras::Jni::is_primitive_type<type>::value> name;
 }
 
 #endif // KANDROIDEXTRAS_JNIPROPERTIES_H
